function read_GHeader(path::String)
    data = open(path, "r")
    g_header = zeros(Int32, 4)
    read!(data, g_header)
    close(data)
    a = GHeader(g_header)
    return a
end

function read_CHeader(path::String; legacy::Bool=false)
    gh = read_GHeader(path)
    
    data = open(path, "r")
    seek(data, gh.hsize)
    
    a = Vector{CHeader}(undef, gh.ncorr)
    if !legacy
        aux_f = zeros(Float64, 6)
        aux_i = zeros(Int32, 4)
        theta = zeros(Float64, 6)

        for k = 1:gh.ncorr
            read!(data, aux_f)
            read!(data, theta)

            qs1 = read(data, Int32)
            if qs1 != 0
                qn1 = read(data, Int32)
                qeps1 = read(data, Float64)
                q1 = Sm(qs1, qn1, qeps1, 1)
            else
                q1 = Sm(qs1, 1)
            end

            qs2 = read(data, Int32)
            if qs2 != 0
                qn2 = read(data, Int32)
                qeps2 = read(data, Float64)
                q2 = Sm(qs2, qn2, qeps2, 1)
            else
                q2 = Sm(qs2, 1)
            end


            gs1 = read(data, Int32)
            if gs1 != 0 && gs1 != 3 && gs1 != 4
                gn1 = read(data, Int32)
                geps1 = read(data, Float64)
                g1 = Sm(gs1, gn1, geps1, 2)
            elseif gs1 == 3 || gs1 == 4
                g1 = Sm(gs1, q1.niter, q1.eps, 2)
            else
                g1 = Sm(gs1, 2)
            end
            
            gs2 = read(data, Int32)
            if gs2 != 0 && gs2 != 3 && gs2 != 4
                gn2 = read(data, Int32)
                geps2 = read(data, Float64)
                g2 = Sm(gs2, gn2, geps2, 2)
            elseif gs1 == 3 || gs1 == 4
                g2 = Sm(gs2, q2.niter, q2.eps, 2)
            else
                g2 = Sm(gs2, 2)
            end
            
            read!(data, aux_i)
            a[k] = CHeader(aux_f, aux_i, theta, [q1, q2, g1, g2])
        end
    else
        aux_f = zeros(Float64, 4)
        aux_i = zeros(Int32, 4)
        for k = 1:gh.ncorr
            read!(data, aux_f)
            read!(data, aux_i)
            a[k] = CHeader(aux_f, aux_i)
        end
    end
    close(data)
    return a
end

@doc raw"""
    read_mesons(path::String, g1::Union{String, Nothing}=nothing, g2::Union{String, Nothing}=nothing; id::Union{String, Nothing}=nothing, legacy::Bool=false)

    read_mesons(path::Vector{String}, g1::Union{String, Nothing}=nothing, g2::Union{String, Nothing}=nothing; id::Union{String, Nothing}=nothing, legacy::Bool=false)

This function read a mesons dat file at a given path and returns a vector of `CData` structures for different masses and Dirac structures.
Dirac structures `g1` and/or `g2` can be passed as string arguments in order to filter correaltors.
ADerrors id can be specified as argument. If is not specified, the `id` is fixed according to the ensemble name (example: "H400"-> id = "H400")

*For the old version (without smearing, distance preconditioning and theta) set legacy=true.

Examples:
```@example
read_mesons(path)
read_mesons(path, "G5")
read_mesons(path, nothing, "G5")
read_mesons(path, "G5", "G5")
read_mesons(path, "G5", "G5", id="H100")
read_mesons(path, "G5_d2", "G5_d2", legacy=true)
```
"""
function read_mesons(path::String, g1::Union{String, Nothing}=nothing, g2::Union{String, Nothing}=nothing; id::Union{String, Nothing}=nothing, legacy::Bool=false)    
    t1 = isnothing(g1) ? nothing : findfirst(x-> x==g1, gamma_name) - 1
    t2 = isnothing(g2) ? nothing : findfirst(x-> x==g2, gamma_name) - 1
    if isnothing(id)
        bname = basename(path)
        m = findfirst(r"[A-Z][0-9]{3}r[0-9]{3}", bname)
        id = bname[m[1:4]]
        #id = parse(Int64, bname[m[2:4]])
    end

    data = open(path, "r")
    g_header = read_GHeader(path)
    c_header = read_CHeader(path, legacy=legacy)

    ncorr = g_header.ncorr
    tvals = g_header.tvals
    nnoise = g_header.nnoise

    fsize = filesize(path)

    datsize = 4 + sum(getfield.(c_header, :dsize)) * tvals * nnoise #data_size / ncnfg
    ncfg = div(fsize - g_header.hsize - sum(getfield.(c_header, :hsize)), datsize) #(total size - header_size) / data_size

    corr_match = findall(x-> (x.type1==t1 || isnothing(t1)) && (x.type2==t2 || isnothing(t2)), c_header)

        
    seek(data, g_header.hsize + sum(c.hsize for c in c_header))

    res = Array{CData}(undef, length(corr_match))

    data_re = Array{Float64}(undef, length(corr_match), ncfg, tvals)
    data_im = zeros(length(corr_match), ncfg, tvals)
    vcfg = Array{Int32}(undef, ncfg)

    for icfg = 1:ncfg
        vcfg[icfg] = read(data, Int32)
        c=1
        for k = 1:ncorr
            if k in corr_match 
                if c_header[k].is_real == 1
                    tmp = Array{Float64}(undef, tvals*nnoise)
                    read!(data, tmp)
                    tmp2 = reshape(tmp, (nnoise, tvals))
                    tmp2 = mean(tmp2, dims=1)
                    data_re[c, icfg, :] = tmp2[1, :]
                elseif c_header[k].is_real == 0
                    tmp = Array{Float64}(undef, 2*tvals*nnoise)
                    read!(data, tmp)
                    tmp2 = reshape(tmp, (2, nnoise, tvals))
                    tmp2 = mean(tmp2, dims=2)
                    data_re[c, icfg, :] = tmp2[1, 1, :]
                    data_im[c, icfg, :] = tmp2[2, 1, :]

                end
                c += 1
            else
                seek(data, position(data)  + c_header[k].dsize*tvals*nnoise)
            end
        

        end
    end

    for c in 1:length(corr_match)
        res[c] = CData(c_header[corr_match[c]], vcfg, data_re[c, :, :], data_im[c, :, :], id)
    end
    close(data)

    return res
end

function read_mesons(path::Vector{String}, g1::Union{String, Nothing}=nothing, g2::Union{String, Nothing}=nothing; id::Union{String, Nothing}=nothing, legacy::Bool=false)
    res = read_mesons.(path, g1, g2, id=id, legacy=legacy)
    nrep = length(res)
    ncorr = length(res[1])

    cdata = Vector{Vector{CData}}(undef, ncorr)
    for icorr = 1:ncorr
        cdata[icorr] = Vector{CData}(undef, nrep) 
        for r = 1:nrep
            cdata[icorr][r] = res[r][icorr]
        end
    end
    return cdata
end

function read_rw(path::String; v::String="1.2")
    data = open(path, "r")
    nrw = read(data, Int32)

    if v == "1.4" || v =="1.6"
        nfct = Array{Int32}(undef, nrw)
        read!(data, nfct)
        nfct_inheader = 1
    elseif v=="1.2"
        nfct = ones(Int32, nrw)
        nfct_inheader = 0
    else
        error("Version not supported")  
    end
    nsrc = Array{Int32}(undef, nrw)
    read!(data, nsrc)
    glob_head_size = 4 + 4*nrw*(nfct_inheader + 1)
    datsize = 4 + 2*8*sum(nsrc .* nfct)

    fsize = filesize(path)

    ncnfg = Int32((fsize - glob_head_size)/datsize)
    r_data = Array{Array{Float64}}(undef, nrw)
    
    [r_data[k] = zeros(Float64, nfct[k], nsrc[k], ncnfg) for k = 1:nrw]
    vcfg = Array{Int32}(undef, ncnfg)
    for icfg = 1:ncnfg
        vcfg[icfg] = read(data, Int32)
        for irw = 1:nrw
            for ifct = 1:nfct[irw]
                tmp = zeros(Float64, nsrc[irw])
                seek(data, position(data) + 8 * nsrc[irw])
                read!(data, tmp)
                r_data[irw][ifct, :, icfg] = tmp[:]
            end
        end
    end
    close(data)
    return r_data
end

@doc raw"""
    read_ms1(path::String; v::String="1.2")

Reads openQCD ms1 dat files at a given path. This method returns a matrix `W[irw, icfg]` that contains the reweighting factors, where
`irw` is the `rwf` index and icfg the configuration number.
The function is compatible with the output files of openQCD v=1.2, 1.4 and 1.6. Version can be specified as argument.

Examples:
```@example    
read_ms1(path)
read_ms1(path, v="1.4")
read_ms1(path, v="1.6")
```
"""
function read_ms1(path::String; v::String="1.2")
    r_data = read_rw(path, v=v)
    nrw = length(r_data)
    ncnfg = size(r_data[1])[3]
    W = zeros(Float64, nrw, ncnfg)
    [W[k, :] = prod(mean(exp.(.-r_data[k]), dims=2), dims=1) for k = 1:nrw]
    return W
end
@doc raw"""
    read_md(path::String)

Reads openQCD  pbp.dat files at a given path. This method returns a matrix `md[irw, icfg]` that contains the derivatives ``dS/dm``, where
``md[irw=1] = dS/dm_l`` and ``md[irw=2] = dS/dm_s``

``Seff = -tr(log(D+m))``

``dSeff/ dm = -tr((D+m)^-1)``

Examples:
```@example    
md = read_md(path)
```
"""
function read_md(path::String)
    r_data = read_rw(path, v="1.4")
    nrw = length(r_data)
    ncnfg = size(r_data[1])[3]
    md = zeros(Float64, nrw, ncnfg)
    [md[k, :] = prod(.-mean(r_data[k], dims=2), dims=1) for k = 1:nrw]
    return md
end

@doc raw"""
    read_ms(path::String; id::Union{String, Nothing}=nothing, dtr::Int64=1, obs::String="Y")  

Reads openQCD ms dat files at a given path. This method return YData: 
    
- `t(t)`: flow time values

- `obs(icfg, x0, t)`: the time-slice sums of the densities of the observable (Wsl, Ysl or Qsl)

- `vtr`: vector that contains trajectory number

- `id`: ensmble id

`dtr` = `dtr_cnfg` / `dtr_ms`, where `dtr_cnfg` is the number of trajectories computed before saving the configuration. `dtr_ms`
is the same but applied to the ms.dat file.

Examples:
```@example    
Y = read_ms(path)
```
"""
function read_ms(path::String; id::Union{String, Nothing}=nothing, dtr::Int64=1 , obs::String="Y")  
    if isnothing(id)
        bname = basename(path)
        m = findfirst(r"[A-Z][0-9]{3}r[0-9]{3}", bname)
        id = bname[m[1:4]]
        #id = parse(Int64, bname[m[2:4]])
    end  
    data = open(path, "r")

    dn = read(data, Int32)
    nn = read(data, Int32)
    tvals = read(data, Int32)
    eps = read(data, Float64)

    fsize = filesize(path)
    datsize=4 + 3*8*(nn + 1) * tvals # measurement size of each cnfg

    ntr = Int32((fsize - 3*4 - 8) / datsize)

    if mod(ntr, dtr) != 0
        error("ntr / dtr must be exact")
    end

    vntr = Vector{Int32}(undef, div(ntr, dtr))
    
    # x0, t, cfg 
    Wsl = Array{Float64}(undef, div(ntr, dtr), tvals, nn + 1) 
    Ysl = Array{Float64}(undef, div(ntr, dtr), tvals, nn + 1)
    Qsl = Array{Float64}(undef, div(ntr, dtr), tvals, nn + 1)

    k = 0
    for itr = 1:ntr
        tmp = read(data, Int32)
        if mod(itr, dtr) == 0
            k += 1
            vntr[k] = tmp
        end

        for iobs = 1:3
            for inn = 0:nn
                tmp2 = Vector{Float64}(undef, tvals)
                read!(data, tmp2)
                if mod(itr, dtr) == 0
                    if iobs == 1
                        Wsl[k, :, inn + 1] = tmp2
                    elseif iobs == 2
                        Ysl[k, :, inn + 1] = tmp2
                    elseif iobs == 3
                        Qsl[k, :, inn + 1] = tmp2
                    end
                end
            end
        end
    end
    close(data)
    t = Float64.(0:nn) .* dn .* eps
    
    if obs == "W"
        return YData(vntr, t, Wsl, id)
    elseif obs == "Y"
        return YData(vntr, t, Ysl, id)
    elseif obs == "Q"
        return YData(vntr, t, Qsl, id)
    else
        println("obs = ", obs," is not valid")
        return nothing
    end
end

@doc raw"""
    truncate_data!(data::YData, nc::Int64)

    truncate_data!(data::Vector{YData}, nc::Vector{Int64})

    truncate_data!(data::Vector{CData}, nc::Int64)

    truncate_data!(data::Vector{Vector{CData}}, nc::Vector{Int64})

Truncates the output of `read_mesons` and `read_ms` taking the first `nc` configurations.

Examples:
```@example    
#Single replica
dat = read_mesons(path, "G5", "G5")
Y = read_ms(path)
truncate_data!(dat, nc)
truncate_data!(Y, nc)

#Two replicas
dat = read_mesons([path1, path2], "G5", "G5")
Y = read_ms.([path1_ms, path2_ms])
truncate_data!(dat, [nc1, nc2])
truncate_data!(Y, [nc1, nc2])
```
"""
function truncate_data!(data::YData, nc::Int64)
    data.vtr = data.vtr[1:nc]
    data.obs = data.obs[1:nc, :, :]
    return nothing
end
function truncate_data!(data::Vector{YData}, nc::Vector{Int64})
    truncate_data!.(data, nc)
    return nothing
end
function truncate_data!(data::Vector{CData}, nc::Int64)
    N = length(data)
    for k = 1:N
        data[k].vcfg = data[k].vcfg[1:nc]
        data[k].re_data = data[k].re_data[1:nc, :]
        data[k].im_data = data[k].im_data[1:nc, :]
    end
    return nothing
end
function truncate_data!(data::Vector{Vector{CData}}, nc::Vector{Int64})
    N = length(data)
    R = length(data[1])
    for k = 1:N
        for r = 1:R
            data[k][r].vcfg = data[k][r].vcfg[1:nc[r]]
            data[k][r].re_data = data[k][r].re_data[1:nc[r], :]
            data[k][r].im_data = data[k][r].im_data[1:nc[r], :]
        end
    end
    return nothing
end